@@ -9,6 +9,7 @@ use crate::{
     core::{
         engine_state::{genesis::GenesisError, upgrade::ProtocolUpgradeError},
         execution,
+        runtime::stack,
     },
     shared::wasm_prep,
     storage,
@@ -78,6 +79,9 @@ pub enum Error {
     /// Missing system contract hash.
     #[error("Missing system contract hash: {0}")]
     MissingSystemContractHash(String),
+    /// An attempt to push to the runtime stack while already at the maximum height.
+    #[error("Runtime stack overflow")]
+    RuntimeStackOverflow,
 }
 
 impl Error {
@@ -120,6 +124,12 @@ impl From<GenesisError> for Error {
     }
 }
 
+impl From<stack::RuntimeStackOverflow> for Error {
+    fn from(_: stack::RuntimeStackOverflow) -> Self {
+        Self::RuntimeStackOverflow
+    }
+}
+
 impl DataSize for Error {
     const IS_DYNAMIC: bool = true;
 
