@@ -1,10 +1,8 @@
-use blake2::{
-    digest::{Update, VariableOutput},
-    VarBlake2b,
-};
+//! Generates unique 32-byte addresses.
 use rand::{RngCore, SeedableRng};
 use rand_chacha::ChaChaRng;
 
+use casper_hashing::Digest;
 use casper_types::{AccessRights, Phase, URef};
 
 use crate::core::{Address, ADDRESS_LENGTH};
@@ -23,23 +21,20 @@ impl AddressGenerator {
             .build()
     }
 
+    /// Creates a new [`Address`] by using an internal instance of PRNG.
     pub fn create_address(&mut self) -> Address {
         let mut buff = [0u8; ADDRESS_LENGTH];
         self.0.fill_bytes(&mut buff);
         buff
     }
 
+    /// Creates a new [`Address`] by hashing an output from [`AddressGenerator::create_address`]
+    /// with a blake2b256.
     pub fn new_hash_address(&mut self) -> Address {
-        // TODO: this appears to duplicate the logic of AddressGeneratorBuilder::build()
-        let pre_hash_bytes = self.create_address();
-        // NOTE: Unwrap below is assumed safe as output size of `ADDRESS_LENGTH` is a valid value.
-        let mut hasher = VarBlake2b::new(ADDRESS_LENGTH).unwrap();
-        hasher.update(&pre_hash_bytes);
-        let mut hash_bytes = [0; ADDRESS_LENGTH];
-        hasher.finalize_variable(|hash| hash_bytes.clone_from_slice(hash));
-        hash_bytes
+        Digest::hash(self.create_address()).value()
     }
 
+    /// Creates a new [`URef`] with a new address generated.
     pub fn new_uref(&mut self, access_rights: AccessRights) -> URef {
         let addr = self.create_address();
         URef::new(addr, access_rights)
@@ -53,21 +48,22 @@ pub struct AddressGeneratorBuilder {
 }
 
 impl AddressGeneratorBuilder {
+    /// Creates a new builder.
     pub fn new() -> Self {
         Default::default()
     }
 
+    /// Extends the seed with more data.
     pub fn seed_with(mut self, bytes: &[u8]) -> Self {
         self.data.extend(bytes);
         self
     }
 
+    /// Creates a new [`AddressGenerator`].
+    ///
+    /// This method hashes the seed bytes, and seeds the PRNG with it.
     pub fn build(self) -> AddressGenerator {
-        let mut seed: [u8; SEED_LENGTH] = [0u8; SEED_LENGTH];
-        // NOTE: Unwrap below is assumed safe as output size of `SEED_LENGTH` is a valid value.
-        let mut hasher = VarBlake2b::new(SEED_LENGTH).unwrap();
-        hasher.update(self.data);
-        hasher.finalize_variable(|hash| seed.clone_from_slice(hash));
+        let seed: [u8; SEED_LENGTH] = Digest::hash(self.data).value();
         AddressGenerator(ChaChaRng::from_seed(seed))
     }
 }
