@@ -16,18 +16,16 @@ use super::{
     meter::count_meter::Count, AddResult, TrackingCopy, TrackingCopyCache, TrackingCopyQueryResult,
 };
 use crate::{
-    core::{
-        engine_state::{op::Op, EngineConfig},
-        runtime_context::dictionary,
-        ValidationError,
-    },
-    shared::{newtypes::CorrelationId, transform::Transform},
+    core::{engine_state::EngineConfig, runtime_context::dictionary, ValidationError},
+    shared::{execution_journal::ExecutionJournal, newtypes::CorrelationId, transform::Transform},
     storage::{
         global_state::{in_memory::InMemoryGlobalState, StateProvider, StateReader},
         trie::merkle_proof::TrieMerkleProof,
     },
 };
 
+const MAX_STORED_VALUE_SIZE: u32 = 8 * 1024 * 1024;
+
 struct CountingDb {
     count: Rc<Cell<i32>>,
     value: Option<StoredValue>,
@@ -88,8 +86,7 @@ fn tracking_copy_new() {
     let db = CountingDb::new(counter);
     let tc = TrackingCopy::new(db);
 
-    assert!(tc.ops.is_empty());
-    assert!(tc.fns.is_empty());
+    assert!(tc.journal.is_empty());
 }
 
 #[test]
@@ -125,12 +122,11 @@ fn tracking_copy_read() {
     let value = tc.read(correlation_id, &k).unwrap().unwrap();
     // value read correctly
     assert_eq!(value, zero);
-    // read produces an identity transform
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::Identity));
-    // read does produce an op
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Read));
+    // Reading does produce an identity transform.
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::Identity)])
+    );
 }
 
 #[test]
@@ -144,25 +140,24 @@ fn tracking_copy_write() {
     let two = StoredValue::CLValue(CLValue::from_t(2_i32).unwrap());
 
     // writing should work
-    tc.write(k, one.clone());
+    let _ = tc.write(k, one.clone(), MAX_STORED_VALUE_SIZE);
     // write does not need to query the DB
     let db_value = counter.get();
     assert_eq!(db_value, 0);
-    // write creates a Transform
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::Write(one)));
-    // write creates an Op
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Write));
+    // Writing creates a write transform.
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::Write(one.clone()))])
+    );
 
     // writing again should update the values
-    tc.write(k, two.clone());
+    let _ = tc.write(k, two.clone(), MAX_STORED_VALUE_SIZE);
     let db_value = counter.get();
     assert_eq!(db_value, 0);
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::Write(two)));
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Write));
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::Write(one)), (k, Transform::Write(two))])
+    );
 }
 
 #[test]
@@ -176,23 +171,22 @@ fn tracking_copy_add_i32() {
     let three = StoredValue::CLValue(CLValue::from_t(3_i32).unwrap());
 
     // adding should work
-    let add = tc.add(correlation_id, k, three.clone());
+    let add = tc.add(correlation_id, k, three.clone(), MAX_STORED_VALUE_SIZE);
     assert_matches!(add, Ok(_));
 
-    // add creates a Transform
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::AddInt32(3)));
-    // add creates an Op
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Add));
+    // Adding creates an add transform.
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::AddInt32(3))])
+    );
 
     // adding again should update the values
-    let add = tc.add(correlation_id, k, three);
+    let add = tc.add(correlation_id, k, three, MAX_STORED_VALUE_SIZE);
     assert_matches!(add, Ok(_));
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::AddInt32(6)));
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Add));
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::AddInt32(3)); 2])
+    );
 }
 
 #[test]
@@ -219,36 +213,40 @@ fn tracking_copy_add_named_key() {
     let name2 = "test2".to_string();
     let other_named_key = StoredValue::CLValue(CLValue::from_t((name2.clone(), u2)).unwrap());
     let mut map = NamedKeys::new();
-    map.insert(name1, u1);
+    map.insert(name1.clone(), u1);
 
     // adding the wrong type should fail
     let failed_add = tc.add(
         correlation_id,
         k,
         StoredValue::CLValue(CLValue::from_t(3_i32).unwrap()),
+        MAX_STORED_VALUE_SIZE,
     );
     assert_matches!(failed_add, Ok(AddResult::TypeMismatch(_)));
-    assert!(tc.ops.is_empty());
-    assert!(tc.fns.is_empty());
+    assert!(tc.journal.is_empty());
 
     // adding correct type works
-    let add = tc.add(correlation_id, k, named_key);
+    let add = tc.add(correlation_id, k, named_key, MAX_STORED_VALUE_SIZE);
     assert_matches!(add, Ok(_));
-    // add creates a Transform
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::AddKeys(map.clone())));
-    // add creates an Op
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Add));
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(
+            k,
+            Transform::AddKeys(iter::once((name1.clone(), u1)).collect())
+        )])
+    );
 
     // adding again updates the values
-    map.insert(name2, u2);
-    let add = tc.add(correlation_id, k, other_named_key);
+    map.insert(name2.clone(), u2);
+    let add = tc.add(correlation_id, k, other_named_key, MAX_STORED_VALUE_SIZE);
     assert_matches!(add, Ok(_));
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::AddKeys(map)));
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Add));
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![
+            (k, Transform::AddKeys(iter::once((name1, u1)).collect())),
+            (k, Transform::AddKeys(iter::once((name2, u2)).collect()))
+        ])
+    );
 }
 
 #[test]
@@ -262,11 +260,11 @@ fn tracking_copy_rw() {
     // reading then writing should update the op
     let value = StoredValue::CLValue(CLValue::from_t(3_i32).unwrap());
     let _ = tc.read(correlation_id, &k);
-    tc.write(k, value.clone());
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::Write(value)));
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Write));
+    let _ = tc.write(k, value.clone(), MAX_STORED_VALUE_SIZE);
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::Identity), (k, Transform::Write(value))])
+    );
 }
 
 #[test]
@@ -280,12 +278,11 @@ fn tracking_copy_ra() {
     // reading then adding should update the op
     let value = StoredValue::CLValue(CLValue::from_t(3_i32).unwrap());
     let _ = tc.read(correlation_id, &k);
-    let _ = tc.add(correlation_id, k, value);
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::AddInt32(3)));
-    assert_eq!(tc.ops.len(), 1);
-    // this Op is correct because Read+Add = Write
-    assert_eq!(tc.ops.get(&k), Some(&Op::Write));
+    let _ = tc.add(correlation_id, k, value, MAX_STORED_VALUE_SIZE);
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![(k, Transform::Identity), (k, Transform::AddInt32(3))])
+    );
 }
 
 #[test]
@@ -299,12 +296,15 @@ fn tracking_copy_aw() {
     // adding then writing should update the op
     let value = StoredValue::CLValue(CLValue::from_t(3_i32).unwrap());
     let write_value = StoredValue::CLValue(CLValue::from_t(7_i32).unwrap());
-    let _ = tc.add(correlation_id, k, value);
-    tc.write(k, write_value.clone());
-    assert_eq!(tc.fns.len(), 1);
-    assert_eq!(tc.fns.get(&k), Some(&Transform::Write(write_value)));
-    assert_eq!(tc.ops.len(), 1);
-    assert_eq!(tc.ops.get(&k), Some(&Op::Write));
+    let _ = tc.add(correlation_id, k, value, MAX_STORED_VALUE_SIZE);
+    let _ = tc.write(k, write_value.clone(), MAX_STORED_VALUE_SIZE);
+    assert_eq!(
+        tc.journal,
+        ExecutionJournal::new(vec![
+            (k, Transform::AddInt32(3)),
+            (k, Transform::Write(write_value))
+        ])
+    );
 }
 
 proptest! {
@@ -934,7 +934,7 @@ fn get_keys_should_return_keys_in_the_uref_keyspace() {
     let cl_value = CLValue::from_t(U512::from(2)).expect("should convert");
     let uref_3_value = StoredValue::CLValue(cl_value);
     let uref_3_key = Key::URef(URef::new([10; 32], AccessRights::READ_ADD_WRITE));
-    tracking_copy.write(uref_3_key, uref_3_value);
+    let _ = tracking_copy.write(uref_3_key, uref_3_value, MAX_STORED_VALUE_SIZE);
 
     let key_set = tracking_copy
         .get_keys(correlation_id, &KeyTag::URef)
@@ -970,7 +970,7 @@ fn get_keys_should_handle_reads_from_empty_trie() {
     let cl_value = CLValue::from_t(U512::zero()).expect("should convert");
     let uref_1_value = StoredValue::CLValue(cl_value);
     let uref_1_key = Key::URef(URef::new([8; 32], AccessRights::READ_ADD_WRITE));
-    tracking_copy.write(uref_1_key, uref_1_value);
+    let _ = tracking_copy.write(uref_1_key, uref_1_value, MAX_STORED_VALUE_SIZE);
 
     let key_set = tracking_copy
         .get_keys(correlation_id, &KeyTag::URef)
@@ -983,7 +983,7 @@ fn get_keys_should_handle_reads_from_empty_trie() {
     let cl_value = CLValue::from_t(U512::one()).expect("should convert");
     let uref_2_value = StoredValue::CLValue(cl_value);
     let uref_2_key = Key::URef(URef::new([9; 32], AccessRights::READ_ADD_WRITE));
-    tracking_copy.write(uref_2_key, uref_2_value);
+    let _ = tracking_copy.write(uref_2_key, uref_2_value, MAX_STORED_VALUE_SIZE);
 
     let key_set = tracking_copy
         .get_keys(correlation_id, &KeyTag::URef)
@@ -1002,7 +1002,7 @@ fn get_keys_should_handle_reads_from_empty_trie() {
         fake_purse,
     ));
     let account_key = Key::Account(account_hash);
-    tracking_copy.write(account_key, account_value);
+    let _ = tracking_copy.write(account_key, account_value, MAX_STORED_VALUE_SIZE);
 
     assert_eq!(key_set.len(), 2);
     assert!(key_set.contains(&uref_1_key.normalize()));
@@ -1013,7 +1013,7 @@ fn get_keys_should_handle_reads_from_empty_trie() {
     let cl_value = CLValue::from_t(U512::from(2)).expect("should convert");
     let uref_3_value = StoredValue::CLValue(cl_value);
     let uref_3_key = Key::URef(URef::new([10; 32], AccessRights::READ_ADD_WRITE));
-    tracking_copy.write(uref_3_key, uref_3_value);
+    let _ = tracking_copy.write(uref_3_key, uref_3_value, MAX_STORED_VALUE_SIZE);
 
     let key_set = tracking_copy
         .get_keys(correlation_id, &KeyTag::URef)
