@@ -1,6 +1,6 @@
 //! The shared logic of the execution engine.
 pub mod additive_map;
-#[macro_use]
+pub mod execution_journal;
 pub mod host_function_costs;
 pub mod logging;
 pub mod newtypes;
@@ -10,6 +10,5 @@ pub mod system_config;
 pub mod test_utils;
 pub mod transform;
 pub mod utils;
-pub mod wasm;
 pub mod wasm_config;
 pub mod wasm_prep;
