@@ -1,8 +1,10 @@
 //! Some functions to use in tests.
 
-use casper_types::{account::AccountHash, contracts::NamedKeys, AccessRights, Key, URef};
-
-use crate::shared::{account::Account, stored_value::StoredValue};
+use casper_types::{
+    account::{Account, AccountHash},
+    contracts::NamedKeys,
+    AccessRights, Key, StoredValue, URef,
+};
 
 /// Returns an account value paired with its key
 pub fn mocked_account(account_hash: AccountHash) -> Vec<(Key, StoredValue)> {
